-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local stringLib = npcg.functua.string
local util 			= npcg.functua.util

-------------------------------------------------------------------------------------
-- | Variable Definitions |----------------------------------------------------------
-------------------------------------------------------------------------------------

-- Define a minimap icon API.
npcg.minimapIcon = { }

-- Active options, these represent the live, non-cache states of the addon.
-- These should be read from the IgnoreEnemyEmotesDB on loading, and flushed to it on unloading.
-- These should be read from the cache on confirming changes.
local savedVars = { }

-- Cached options.
-- These should be flushed to the savedVars on confirming changes.
-- These should be cleared when leaving the options menu.
local cache       = { }

-- Default options. These cannot be changed after being set.
-- These must be explicitly defined for every savedVar.
local defaults    = {
  ["minimapPosition"  ] = 45, -- On Minimap
  ["minimapPositionX" ] = 0,  -- Not on the Minimap
  ["minimapPositionY" ] = 0,  -- Not on the Minimap
}

-------------------------------------------------------------------------------------
-- | Helper Functions |--------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Helper function to convert the minimapPosition variable to a concrete x and y position.
local function getPositionOnMinimap()
  return 52-(80*cos(npcg.options.savedVars["minimapPosition"])),(80*sin(npcg.options.savedVars["minimapPosition"]))-52
end


local function repositionMinimapButton()
	npcg.options.savedVars["minimapPosition" ] = npcg.options.savedVars["minimapPosition" ] or defaults["minimapPosition"]
	npcg.options.savedVars["minimapPositionX"] = npcg.options.savedVars["minimapPositionX"] or npcGenerator_Minimap_Button:GetLeft() 
	npcg.options.savedVars["minimapPositionY"] = npcg.options.savedVars["minimapPositionY"] or npcGenerator_Minimap_Button:GetTop()
	
	npcGenerator_Minimap_Button:ClearAllPoints()
	
	if npcg.options.savedVars["lockToMinimap"] then
		npcGenerator_Minimap_Button:SetPoint("TOPLEFT", "Minimap",  "TOPLEFT", getPositionOnMinimap())
	else
		npcGenerator_Minimap_Button:SetPoint("CENTER", "UIParent", "BOTTOMLEFT", npcg.options.savedVars["minimapPositionX"], npcg.options.savedVars["minimapPositionY"])
	end
end


function npcg.minimapIcon.Reset(self)
	npcg.options.savedVars["minimapPosition"] = defaults["minimapPosition"]
	npcGenerator_Minimap_Button:SetPoint("TOPLEFT","Minimap","TOPLEFT", getPositionOnMinimap())
end

function npcg.minimapIcon.Invalidate(self)
  if   npcg.options.savedVars["displayMinimapIcon"]
  then npcGenerator_Minimap_Button:Show()
  else npcGenerator_Minimap_Button:Hide() end
  
  repositionMinimapButton()
end


local function npcGenerator_minimapButton_WhileDragging()
	local xpos,ypos = GetCursorPosition()
	
	if npcg.options.savedVars["lockToMinimap"] then
		local xmin,ymin = Minimap:GetLeft(), Minimap:GetBottom()

		xpos = xmin-xpos/UIParent:GetScale()+70 -- get coordinates as differences from the center of the minimap
		ypos = ypos/UIParent:GetScale()-ymin-70

		npcg.options.savedVars["minimapPosition" ] = math.deg(math.atan2(ypos,xpos)) -- save the degrees we are relative to the minimap center
	else
		npcg.options.savedVars["minimapPositionX"] = xpos / UIParent:GetEffectiveScale()
		npcg.options.savedVars["minimapPositionY"] = ypos / UIParent:GetEffectiveScale()
	end
	repositionMinimapButton() -- move the button
end

-------------------------------------------------------------------------------------
-- | Event Hooks and Handlers |------------------------------------------------------
-------------------------------------------------------------------------------------

-------------------------------------------------------------------------------------
---- | Loading and Unloading |-------------------------------------------------------
-------------------------------------------------------------------------------------

-- Loads all options from external storage to NPCG_Options.
-- To be called on ADDON_LOADED.
function npcg.loadMinimapIcon()
  savedVars = npcg.options.savedVars
  
  -- Register Minimap for movement.
  npcGenerator_Minimap_Button:RegisterForClicks(  "LeftButtonUp", "RightButtonUp")
	npcGenerator_Minimap_Button:RegisterForDrag(      "LeftButton",   "RightButton")
  
  repositionMinimapButton()
	if npcg.options.savedVars["displayMinimapIcon"] then npcGenerator_Minimap_Button:Show()
    else npcGenerator_Minimap_Button:Hide() end
    
  npcg.minimapIcon:Invalidate()
  
  -- Set Tooltip script.
  npcGenerator_Minimap_Button:SetScript("OnEnter", function(self)
  	GameTooltip:SetOwner(npcGenerator_Minimap_Button, "ANCHOR_LEFT")
    GameTooltip:SetText("npcGenerator", 1, 1, 1)
    GameTooltip:AddLine("Open the npcGenerator window", 1.00,	0.82,	0.00, 1)
    GameTooltip:Show()
    GameTooltip:FadeOut() end)
  
  -- Set OnClick script.
  npcGenerator_Minimap_Button:SetScript("OnClick", function(self)
    local frame = npcGenerator_NPC_Frame
    if frame:IsVisible() then
      frame:Hide()
    else
      frame:Show()
    end end)
  
  -- Set Dragging Scripts
  npcGenerator_Minimap_Button:SetScript("OnDragStart", function(self)
  	npcGenerator_Minimap_Button:LockHighlight()
    npcGenerator_Minimap_Button_DraggingFrame:Show() end)
  npcGenerator_Minimap_Button:SetScript("OnDragStop",  function(self)
  	npcGenerator_Minimap_Button:UnlockHighlight()
    npcGenerator_Minimap_Button_DraggingFrame:Hide() end)
   
  npcGenerator_Minimap_Button_DraggingFrame:SetScript("OnUpdate", npcGenerator_minimapButton_WhileDragging)
end

-- eof