-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua   = npcg.functua
local stringLib = npcg.functua.string
local util      = npcg.functua.util

-------------------------------------------------------------------------------------
-- | Variable Definitions |----------------------------------------------------------
-------------------------------------------------------------------------------------

-- Active options, these represent the live, non-cache states of the addon.
-- These should be read from the IgnoreEnemyEmotesDB on loading, and flushed to it on unloading.
-- These should be read from the cache on confirming changes.
local savedVars = { }

-- Cached options.
-- These should be flushed to the savedVars on confirming changes.
-- These should be cleared when leaving the options menu.
local cache     = { }

-- Default options. These cannot be changed after being set.
-- These must be explicitly defined for every savedVar.
local defaults  = {
  ["strictFactionChecking"] =  true,
  ["contextChecking"      ] =  true,
  ["useZones"             ] =  true,
  ["useSubZones"          ] =  true,
  ["useMockZone"          ] = false,
  ["useMockSubZone"       ] = false,
  ["haveFriendsAppear"    ] = false,
  ["mockZone"             ] = "",
  ["mockSubZone"          ] = "",
  ["friendSpawnChance"    ] = "0.1",
}

-------------------------------------------------------------------------------------
-- | Helper Functions |--------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Set up a Checkbox, setting it's description, links it to a savedVar, set up the cache var, handles color changes based on enabling/disabling and sets up the tooltips.
local function setUpCheckboxHelper(checkbox, savedVarName, descriptionText, tooltipWhenChecked, tooltipWhenNotChecked)
  -- Set the name for the description.
  getglobal(checkbox:GetName() .. "Text"):SetText(descriptionText)
  -- On Show retrieve checked status from savedVars.
  checkbox:SetScript(   "OnShow", function(self) self:SetChecked(savedVars[savedVarName]) end)
  -- On Hide clear cache.
  checkbox:SetScript(   "OnHide", function(self) cache[savedVarName] = nil end)
  -- Post Click set cache, not value!
  checkbox:SetScript("PostClick", function(self) cache[savedVarName] = self:GetChecked() end)
  checkbox:SetScript(  "OnClick", function(self) end) -- Defined because Blizz code will throw an error otherwise.
  -- Tooltip
  checkbox:SetScript(  "OnEnter", function(self) SetHelperTextOption(self:GetChecked() and tooltipWhenNotChecked or tooltipWhenChecked, self) end)
  -- Fade text based on checked status.
  checkbox:SetScript( "OnEnable", function(self) getglobal(self:GetName() .. "Text"):SetTextColor(1.0, 1.0, 1.0) end)
  checkbox:SetScript("OnDisable", function(self) getglobal(self:GetName() .. "Text"):SetTextColor(0.5, 0.5, 0.5) end)
end


-- Set up an Editbox, linking it to a savedVar, setting it's default text, providing a function to detect invalid input and corrects it, and sets up tooltips.
local function setUpEditboxHelper(editbox, savedVarName, defaultText, isValidText, tooltipText)
  
  -- On Show retrieve checked status from savedVars.
  editbox:SetScript(   "OnShow", function(self)
    self:SetText(isValidText(savedVars[savedVarName]) and savedVars[savedVarName] or defaultText) end)
  -- On Hide clear cache.
  editbox:SetScript(   "OnHide", function(self) cache[savedVarName] = nil end)

  -- Tooltip
  editbox:SetScript(  "OnEnter", function(self) SetHelperTextOption(tooltipText, self) end)
  -- Fade text based on checked status.
  editbox:SetScript( "OnEnable", function(self) self:SetTextColor(0.75, 0.75, 0.75) end)
  editbox:SetScript("OnDisable", function(self) self:SetTextColor(0.50, 0.50, 0.50) end)
  
  -- Parsing the input
  editbox:SetScript("OnEditFocusGained", function(self)
    self:SetTextColor(1.00, 1.00, 1.00)
    if self:GetText() == defaultText then self:SetText("") end
  end)
  editbox:SetScript("OnEditFocusLost", function(self)
    if not isValidText(self:GetText()) then self:SetText(defaultText) end
    self:SetTextColor(0.75, 0.75, 0.75)
    cache[savedVarName] = self:GetText() end)
end

-- Makes the enabled state of a frame depend on a given checkbox.
local function dependencyHelper(frame, dependency)
  
  -- Refresh function
  local refresh = function(dep)
    if    dep:IsEnabled() and dep:GetChecked() 
    then  frame:Enable()
    else  frame:Disable() end
  end
  
  dependency:HookScript("PostClick", function() refresh(dependency) end)
  dependency:HookScript( "OnEnable", function() refresh(dependency) end)
  dependency:HookScript("OnDisable", function() refresh(dependency) end)
  
  frame:HookScript("OnShow", function() refresh(dependency) end)
end

-------------------------------------------------------------------------------------
-- | Event Hooks and Handlers |------------------------------------------------------
-------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------
---- | Loading and Unloading |-------------------------------------------------------
-------------------------------------------------------------------------------------

-- Loads all options from external storage to npcg.options.savedVars.
-- To be called on ADDON_LOADED.
function npcg.loadGeneralPanelOptions(optionsPanelParent)
  savedVars = npcg.options.savedVars

  -- Load savedVars used in this panel.
  functua.table.forEachKey(
    function(k) savedVars[k] = savedVars[k] or defaults[k] end,
    defaults)

  -- Set up the panel.
  local panel                                   = npcGenerator_Options_General_Panel
	panel.name 		                                = "General"
	panel.parent 	                                = optionsPanelParent.name
  
  -- Set up okay button handler.
  panel.okay = function()   
    functua.table.forEachKey(
      function(k)
        savedVars[k]  = cache[k]
        cache[k]      = nil end,
      cache)
    
    -- Maybe.. fire an event instead of directly calling npcg.system.
    npcg.system.remakeGenerator() -- Defined in npcGenerator.lua. Name subject to change to npcg.system.remakeGenerators()
  end  
  
  -- Set up default button handler.
  panel.default = function()
    functua.table.forEachKey(
      function(k) savedVars[k] = defaults[k] end,
      defaults)
  end
  
  -- Set up the cancel handler. This will clear the cache.
  panel.cancel = function()
    functua.table.forEachKey(
      function(k)
        cache[k] = nil end,
      cache)
  end
  
  -- Set up the checkboxes, argument 1 is the checkbox, argument 2 is their related savedVar.
  setUpCheckboxHelper(panel.CheckFaction,     "strictFactionChecking",  "Ban Opposing Faction", "Check to allow only your faction to appear.", "Uncheck to allow all factions to appear.")
  setUpCheckboxHelper(panel.CheckContext,     "contextChecking",        "Make exceptions based on context", "Check to make exceptions based on location or other factors.", "Disable to not make exceptions based on location or other factors.")
  setUpCheckboxHelper(panel.CheckUseZones,    "useZones",               "Use zones in generation", "Check to involve the location in the generation.", "Uncheck to not involve the location in the generation.")
  setUpCheckboxHelper(panel.CheckMockZone,    "useMockZone",            "Use stand-in zone:", "Check to use a location specified instead of the in-game location.", "Check to use the in-game location and not one specified.")
  setUpCheckboxHelper(panel.CheckSubZone,     "useSubZones",            "Use Subzones", "Check to involve the subzone in the generation.", "Uncheck to not involve the subzone in the generation.")
  setUpCheckboxHelper(panel.CheckMockSubZone, "useMockSubZone",         "Use stand-in:", "Check to use a sub location specified instead of the in-game location.", "Uncheck to use the in-game location and not one specified.")
  setUpCheckboxHelper(panel.CheckUseFriends,  "haveFriendsAppear",      "Allow saved npcs to sometimes appear when generating a random npc", "Check to allow saved npcs to appear through random generation.", "Uncheck to not allow saved npcs to appear through random generation.")
  
  -- Set up the editboxes, argument 1 is the checkbox, argument 2 is their related savedVar. Argument 4 is a function that returns whether given input text is valid.
  setUpEditboxHelper(panel.InputMockZone,     "mockZone",               "Enter location here.", function(s) return not stringLib.nilOrWhitespace(s) end, "Enter a location to use instead of the in-game location.")
  setUpEditboxHelper(panel.InputMockSubZone,  "mockSubZone",            "Enter location here.", function(s) return not stringLib.nilOrWhitespace(s) end, "Enter a location to use instead of the in-game location.")
  setUpEditboxHelper(panel.InputChance,       "friendSpawnChance",      "0.1", function(s) return tonumber(s) and tonumber(s) <= 1.0 and tonumber(s) >= 0.0 end, "The chance in range [0,1] a saved npc will appear when generating an npc.")
  
  -- Disables a child element when their parent is unchecked or disabled.
  dependencyHelper(panel.CheckContext,      panel.CheckFaction)
  dependencyHelper(panel.CheckMockZone,     panel.CheckUseZones)
  dependencyHelper(panel.CheckSubZone,      panel.CheckUseZones)
  dependencyHelper(panel.CheckMockSubZone,  panel.CheckSubZone)
  dependencyHelper(panel.InputMockZone,     panel.CheckMockZone)
  dependencyHelper(panel.InputMockSubZone,  panel.CheckMockSubZone)
  dependencyHelper(panel.InputChance,       panel.CheckUseFriends)
  
  InterfaceOptions_AddCategory(panel)
end

-- eof