-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local stringLib = npcg.functua.string
local util 			= npcg.functua.util

-------------------------------------------------------------------------------------
-- | Constants | --------------------------------------------------------------------
-------------------------------------------------------------------------------------

local MAX_TABS = 11

-------------------------------------------------------------------------------------
-- | Events | -----------------------------------------------------------------------
-------------------------------------------------------------------------------------

local function CreateEvent()
	return {
		-- Functions to call when this event is raised.
		Listeners = { },
		-- Raise the event.
		-- Invoke :: Event -> ... -> void
		Invoke = function(self, ...)
			local args = {...}
			functua.list.forEach( -- a, xs
				function(listener) listener(sender, listenerObject, unpack(args)) end,
				self.Listeners) end,
		-- Add a listener to this event.
		-- AddListener :: Event -> (Event, a, ...) -> void
		AddListener = function(self, listenerFunction)
			table.insert(self.Listeners, listenerFunction)
		end,
	}
end

local OnTabChanged 	= CreateEvent()
local OnTabAdded 		= CreateEvent()
local OnTabRemoved 	= CreateEvent()

-------------------------------------------------------------------------------------
-- | Helper Functions |--------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Returns the current number of open tabs.
local function GetTabCount(tabContainer)
	return table.getn(tabContainer.tabButtons)
end

-- Returns the currently opened tab.
local function GetCurrentTab(tabContainer)
	return tabContainer.tabButtons[tabContainer.currentTab]
end

-- Returns the rightmost opened tab.
local function GetLastTab(tabContainer)
	return tabContainer.tabButtons[GetTabCount(tabContainer)]
end

local function InitializeTab(self, tabNumber)
	-- self is the Tab itself.

	self:SetScript("OnClick", function(self, button, isDown)
		if button == "LeftButton" and isDown then
			self:GetParent():SetTab(tabNumber)
		end
	end)
	

end

local function AddTab(tabContainer)
	-- If below maxTabs
	if GetTabCount(tabContainer) < tabContainer.maxTabs then
		-- Raise Tab Count
		local tabCount = GetTabCount(tabContainer) + 1
		-- Create new Tab
		local newTab = CreateFrame("Button", nil, tabContainer, "npcGTabButtonTemplate")
		-- Initialize the new Tab.
		newTab:SetPoint("LEFT", GetLastTab(tabContainer), "RIGHT", 5, 0) -- Move to InitializeTab?
		newTab:SetText(tabCount) -- Move to InitializeTab
		
		table.insert(tabContainer.tabButtons, newTab)
		tabContainer.AddTabButton:SetPoint("LEFT", newTab, "RIGHT", 5, 0)
		InitializeTab(newTab, tabCount)

		-- Raise event OnTabAdded
		OnTabAdded:Invoke(tabCount)
		
		-- Set the new Tab
		tabContainer:SetTab(tabCount)
		
		if GetTabCount(tabContainer) == tabContainer.maxTabs then
			tabContainer.AddTabButton:Disable()
		end
	else
		error("Max Tabs is Reached!")
	end
end

local function SetTab(tabContainer, index)
	-- Why this if? Is it because AddTab scalls this?
	if tabContainer.currentTab <= GetTabCount(tabContainer) then
		-- Stop marking the old tab as active.
		GetCurrentTab(tabContainer):UnlockHighlight()
	end
	tabContainer.currentTab = index
	GetCurrentTab(tabContainer):LockHighlight()
	OnTabChanged:Invoke(index)
	--tabContainer.OnTabChanged(index) -- Rework this one to more proper events.
end

local function RemoveTab(tabContainer, index)
	-- The first Tab cannot be removed.
	if GetTabCount(tabContainer) > 1 then
		local tab = table.remove(tabContainer.tabButtons, GetTabCount(tabContainer))
		tab:Hide()
		tabContainer.AddTabButton:SetPoint("LEFT", GetLastTab(tabContainer), "RIGHT", 5, 0)
		
	-- Raise event.
	OnTabRemoved:Invoke(index)
	--tabContainer.OnTabRemoved(index) -- Rework this one to more proper events.
	end
	
	-- Set the tab to the new current tab.
	if index == 1 then	SetTab(tabContainer, 1)
	else								SetTab(tabContainer, index - 1)
	end
	
	-- Enable the creation of new Tabs if the number is below maxTabs, which should be always anyway.
	if GetTabCount(tabContainer) < tabContainer.maxTabs then
		tabContainer.AddTabButton:Enable()
	end
end

function SimpleTabsTemplate_OnLoad(tabContainer)
	-- Variables
	tabContainer[   "maxTabs"] = MAX_TABS
	tabContainer["tabButtons"] = { tabContainer["TabButton1"] }
	tabContainer["currentTab"] = 1

	
	tabContainer.RemoveTab 	= RemoveTab
	tabContainer.AddTab 		= AddTab
	tabContainer.SetTab 		= SetTab
	
	tabContainer.OnTabAdded 	= OnTabAdded
	tabContainer.OnTabChanged = OnTabChanged
	tabContainer.OnTabRemoved = OnTabRemoved
	-- Lock highlight to denote active tab.
	tabContainer["tabButtons"][1]:LockHighlight()
	InitializeTab(tabContainer["tabButtons"][1], 1)
	
	--tabContainer["AddTabButton"]:RegisterForClicks("AnyDown")
	tabContainer["AddTabButton"]:SetScript("OnClick", function(self, button, isDown)
		if button == "LeftButton" and isDown then
			tabContainer:AddTab()
		end
	end)
	

end