-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

-- Import Modules.
local functua   = npcg.functua
local util      = npcg.functua.util

-- Set up Modules.
local generator = { }
local data      = { }
local tests     = { }

-- Make Modules public in npcg.
npcg.generators.trait = generator
npcg.data.trait       = data
npcg.tests.trait      = tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- type Trait             :: String
-- type TraitConstraints  :: { weight :: Int, tags :: List String, }

-- defaultDataTable :: { Trait, TraitConstraints }
-- Raw default data table. This is imported from the spreadsheet and can be considered as static, thus it should not be altered.
local defaultDataTable = { }

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable

-- default :: { Trait, TraitConstraints } -> { Trait, TraitConstraints } -> { Trait, TraitConstraints }
-- Returns the default trait distrubution template altered by entries within the given table.
local function default(dataTable, replacements)
	local replacements = replacements or { }
	
	return functua.table.mapKV(
		function(k, v)
      -- Either the replacement if it exists, or the original.
			return replacements[k] or functua.table.clone(v) end,
		dataTable) -- DefaultDataTable.
end

-- defaultFromWeightTable :: { Trait, TraitConstraints }{ Trait, Int } -> { Trait, TraitConstraints }
-- Returns the default trait distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(dataTable, replacements)
	local output = default(dataTable, { })
	
  -- For each item in the replacement set, set the weight in the dataTable to the replacement value.
	for k, weight in pairs(replacements) do
		if output[k] ~= nil then -- If output[k] == nil then something is wrong with the replacement data.
			output[k].weight = weight
		end
	end
	
	return output
end

-- fromWeightTable :: { Trait, TraitConstraints } -> { Trait, Int } -> { Trait, TraitConstraints }
-- Returns the default trait distribution template altered by a weight table. The default weight is 0.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function fromWeightTable(dataTable, replacements)
	local output = default(dataTable, { })
	
  -- Set all weights to 0.
	output = functua.table.map(
		function(v) local x = functua.table.clone(v); x.weight = 0; return x end, -- Must clone, v is a table. Is this true...? Default() already clones.
		output)
	
  -- For each item in the replacement set, set the weight in the dataTable to the replacement value.
	for k, weight in pairs(replacements) do
		if output[k] ~= nil then -- If output[k] == nil then something is wrong with the replacement data.
			output[k].weight = weight
		end
	end
	
	return output
end

-- addTag :: { Trait, TraitConstraints } -> String -> { Trait, TraitConstraints }
-- Adds a given tag to every TraitConstraints entry in the dataTable.
local function addTag(dataTable, tag)
	return functua.table.map(
		function(constraints)
			local output = functua.table.clone(constraints) -- Clone to ensure purity.
			output.tags = functua.list.insert(output.tags, tag)
			return output	end,
		dataTable)
end

-- addTag could be defined using addToConstraint

-- addToConstraint :: { Trait, TraitConstraints } -> String -> String -> { Trait, TraitConstraints }
-- Adds a given value to a given dependency.
local function addToConstraint(dataTable, dependency, value)
	return functua.table.map(
		function(constraints)
			local output = functua.table.clone(constraints) -- Clone to ensure purity.
			output[dependency] = functua.list.insert(output[dependency], value)
      
			return output
			end,
		dataTable)
end

-- neutral :: { Trait, TraitConstraints } -> { Trait, TraitConstraints }
-- Adds "neutral" to the traitTypes of all entries in a dataTable.
local function neutral(dataTable)
  return addToConstraint(dataTable, "traitTypes", "neutral")
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These data template can be modified.

-- data :: { Trait, TraitConstraints }
-- The default trait distribution template, this is called when there are no specifics for zones and/or subzones.
local data        = default(defaultDataTable, { })

-- zoneData :: { Zone, { Trait, TraitConstraints } }
-- The trait distribution for specific zones.
local zoneData    = { }

-- subZoneData :: { Zone, { SubZone, { Trait, TraitConstraints } } }
-- The trait distribution for specific subzones.
local subZoneData = { }

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Type GenerationFunction :: DataTable -> ZoneDataTable -> SubZoneDataTable -> CharacterTemplate -> Zone -> SubZone -> DataTable
-- A generation function provides a DataTable based on the Zone, SubZone and CharacterTemplate, selected from the inputs.

-- checkAllLocations :: GenerationFunction
-- Returns a DataTable based on location settings.
local function checkAllLocations(defaultDataTable, zoneDataTable, subZoneDataTable, characterTemplate, zone, subZone)
  return 
    -- Test whether the SubZone has values.
    ((subZoneDataTable[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zoneDataTable[zone]
    -- Fallback to default.
    or defaultDataTable
    -- Fallback in case some extreme breaking happened.
    or { }
end

-- checkZonesOnly :: GenerationFunction
-- Returns a DataTable based on zone, but not subzone settings.
local function checkZonesOnly(defaultDataTable, zoneDataTable, subZoneDataTable, characterTemplate, zone)
  return
    -- Test whether the Zone has values.
    zoneDataTable[zone]
    -- Fallback to default.
    or defaultDataTable
    -- Fallback in case some extreme breaking happened.
    or { }
end

-- dontCheckLocations :: GenerationFunction
-- Returns a DataTable without checking locations.
local function dontCheckLocations(defaultDataTable, zoneDataTable, subZoneDataTable, characterTemplate)
  return
    -- Fallback to default.
    defaultDataTable
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Type FilterFunction :: CharacterTemplate -> TraitConstraints -> Zone -> SubZone -> TraitConstraints -> Bool
-- A filter function reduces the available data entries to only a set of feasible entries.
-- Filter functions might be custom based on the trait.

-- isClassicTrait :: CharacterTemplate -> TraitConstraints -> bool
-- Returns whether a trait should occur on Classic.
local function isClassicTrait(characterTemplate, trait)
  return functua.list.contains("classic", trait.traitTypes)
end

-- isTbcTrait :: CharacterTemplate -> TraitConstraints -> bool
-- Returns whether a trait should occur on Classic TBC.
local function isTbcTrait(characterTemplate, trait)
  return functua.list.contains("tbc", trait.traitTypes)
end

-- isRetailTrait :: CharacterTemplate -> TraitConstraints -> bool
-- Returns whether a trait should occur on Retail.
local function isRetailTrait(characterTemplate, trait)
  return functua.list.contains("retail", trait.traitTypes)
end

-- isNextExpansionTrait :: CharacterTemplate -> TraitConstraints -> bool
-- Returns whether a trait should occur for the next expansion. (For testing purposes.)
local function isNextExpansionTrait(characterTemplate, trait)
  return functua.list.contains("future", trait.traitTypes)
end

-- isTraitOfPlayerFaction :: CharacterTemplate -> TraitConstraints -> bool
-- Returns whether a trait occurs for the player's faction. (Either it is of the same faction, or is neutral.) 
-- Uses UnitFactionGroup("player") to determine the playerfaction.
local function isTraitOfPlayerFaction(characterTemplate, trait)
	return
    functua.table.any(
      function(traitFaction) return traitFaction == string.lower(UnitFactionGroup("player")) or traitFaction == "neutral" end,
      trait.traitTypes)
end

-- isTraitWithinContext :: CharacterTemplate -> TraitConstraints -> Zone -> SubZone -> bool
-- Returns whether a trait is within context.
-- Uses a wider range of constraints than just faction.
local function isTraitWithinContext(characterTemplate, trait, zone, subZone)
  return isTraitOfPlayerFaction(trait)
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- { Trait, TraitConstraints } -> Trait
-- This might change in the future.

local function selectUsingWeightedRandom(selectionList)
  return functua.random.weightedRandomFrom(functua.table.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Queries | ------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- All Queries should return a list of Traits.

-- allTraitsExcept :: { Trait, TraitConstraints } -> { Trait } -> { Trait }
-- Return a list of all traits except the ones passed in.
local function allTraitsExcept(dataTable, exclusions)
  return functua.table.without(functua.table.keys(dataTable), exclusions)
end

-- traitsByType :: { Trait, TraitConstraints } -> String -> { Trait }
-- Returns a list of all traits that have a given traitType.
-- Passing in "any" as the traitType argument will return { }.
local function traitsByType(dataTable, traitType)
  return functua.table.filterKeysByValue(
    function(x) return
      functua.table.contains(traitType, x.traitTypes) end,
      util.eitherOr(traitType == "any", { }, dataTable))
end

-- tagsByType :: { Trait, TraitConstraints } -> String -> { Trait }
-- Returns a list of all traits that have a given tag.
local function traitsByTag(dataTable, tag)
  return functua.table.filterKeysByValue(
    function(x) return
      functua.table.contains(tag, x.tags) end,
      dataTable)
end

-- traitsByConstraint :: { Trait, TraitConstraints } -> String -> String -> { Trait }
-- Returns a list of traits of which a given dependency has a given value.
-- i.e. (_, "races", "tauren") returns all traits which occur when the race is "tauren".
local function traitsByConstraint(dataTable, dependency, value)
  return functua.table.filterKeysByValue(
    function(x) return
      functua.table.contains(value, x[dependency]) end,
      dataTable)
end

-- getDependencies :: { Trait, TraitConstraints } -> { String }
-- Returns a list of traits this trait depends on.
local function getDependencies(dataTable)
  return functua.table.filter(
    function(constraintKey)
      return
        constraintKey ~= "tags"       and
        constraintKey ~= "weight"     and
        constraintKey ~= "traitTypes" end,
    functua.table.keys(functua.table.values(dataTable)))      
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- createRaceGenerator :: DataTable -> ZoneDataTable -> SubZoneDataTable -> 
--                        GeneratorFunction -> FilterFunction -> SelectorFunction ->
--                        (CharacterTemplate -> Zone -> SubZone -> Trait)
-- Creates a trait generator based on given input tables, a generation function, a filter, and a selection function.
-- Returns a function like "generateRandomTrait(characterTemplate, zone, subzone)" which returns a trait (result) as a string.
local function createTraitGenerator(defaultDataTable, zoneDataTable, subZoneDataTable, generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone) 
    return selectorF(
			functua.table.filterV(
				function(trait) return filterF(characterTemplate, trait, zone, subZone) end,
				generatorF(defaultDataTable, zoneData, subZoneData, characterTemplate, zone, subZone))) or "Error: No options found!"
  end
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Default Arguments | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- getDefaultGeneratorFunction :: () -> GeneratorFunction
-- Returns the default generatorF argument for createTraitGenerator based on the current options set in NPCG_Options.locationSettings
-- This function checks the location settings specified in NPCG_Options.locationSettings.
local function getDefaultGeneratorFunction()
  if     not npcg.options.savedVars["useZones"] then
    return dontCheckLocations
  elseif not npcg.options.savedVars["useSubZones"] then
    return checkZonesOnly
  else
    return checkAllLocations
  end
end

-- getDefaultFilterFunction :: () -> FilterFunction
-- Returns the filterF argument for createTraitGenerator based on the current options set in NPCG_Options.factionSettings and the Game Level.
-- This function checks the faction settings specified in NPCG_Options.factionSettings and the Game/Expansion Level.
local function getDefaultFilterFunction()
  local output = functua.const(true)
  
  -- Check faction settings.
  error("NpcG Error: Outdated options format")
  if      not npcg.options.savedVars["strictFactionMatching"] then
    output = functua.const(true)
  elseif  not npcg.options.savedVars["contextChecking"] then
    output = isTraitOfPlayerFaction
  else
    output = isTraitWithinContext
  end
  
  -- Check Expansion Level.
  return functua.bothL(output, util.eitherOrByExpansionLevel(isClassicTrait, isRetailTrait, isNextExpansionTrait))
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Exposing Functions and Data | -----------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

generator.generatorFunctions  = { }
generator.filterFunctions     = { }
generator.selectorFunctions   = { }

generator.generatorFunctions.getDefaultGeneratorFunction  = getDefaultGeneratorFunction
generator.filterFunctions.getDefaultFilterFunction        = getDefaultFilterFunction
generator.selectorFunctions.selectUsingWeightedRandom     = selectUsingWeightedRandom

generator.createTraitGenerator = createTraitGenerator

-- End of File.